package org.greenstone.gatherer.metadata;


import java.io.*;
import java.util.*;
import org.greenstone.gatherer.util.XMLTools;
import org.w3c.dom.*;


/** This class represents one metadata set */
public class MetadataSet
{
    private File metadata_set_file = null;
    private ArrayList metadata_set_elements = new ArrayList();
    private String metadata_set_namespace = null;


    public MetadataSet(File metadata_set_file)
    {
	this.metadata_set_file = metadata_set_file;

	// Parse the specified metadata set file
	Document document = XMLTools.parseXMLFile(metadata_set_file);
	if (document == null) {
	    System.err.println("Error: Could not parse metadata set file " + metadata_set_file.getAbsolutePath());
	    return;
	}

	// Read the metadata set's namespace
	metadata_set_namespace = document.getDocumentElement().getAttribute("namespace");

	// Read the Element elements in the metadata set
	NodeList element_elements_nodelist = document.getElementsByTagName("Element");
	for (int i = 0; i < element_elements_nodelist.getLength(); i++) {
	    Element metadata_element_element = (Element) element_elements_nodelist.item(i);
	    String metadata_element_name = metadata_element_element.getAttribute("name");
	    String metadata_element_name_full = metadata_set_namespace + MetadataTools.NAMESPACE_SEPARATOR + metadata_element_name;
	    metadata_set_elements.add(new MetadataElement(metadata_element_name_full, metadata_element_element));
	}
    }


    public int compareMetadataElements(MetadataElement metadata_element_a, MetadataElement metadata_element_b)
    {
	// Extracted metadata is sorted alphabetically
	if (metadata_set_namespace.equals(MetadataSetManager.EXTRACTED_METADATA_NAMESPACE)) {
	    return metadata_element_a.getName().compareTo(metadata_element_b.getName());
	}

	// Elements in metadata sets are sorted according to the other they are specified in the .mds file
	return metadata_set_elements.indexOf(metadata_element_a) - metadata_set_elements.indexOf(metadata_element_b);
    }


    public String getAttribute(String attribute_name, String language_code)
    {
	// Parse the specified metadata set file
	Document document = XMLTools.parseXMLFile(metadata_set_file);
	if (document == null) {
	    System.err.println("Error: Could not parse metadata set file " + metadata_set_file.getAbsolutePath());
	    return null;
	}

	Element document_element = document.getDocumentElement();
	Element parent_element = document_element;
	boolean language_matches = false;

	NodeList setlanguage_elements_nodelist = document_element.getElementsByTagName("SetLanguage");
	for (int i = 0; i < setlanguage_elements_nodelist.getLength(); i++) {
	    Element current_setlanguage_element = (Element) setlanguage_elements_nodelist.item(i);
	    String current_setlanguage_element_code = current_setlanguage_element.getAttribute("code");
	    if (current_setlanguage_element_code.equals(language_code)) {
		// Found language; use it for the next step
		parent_element = current_setlanguage_element;
		language_matches = true;
		break;
	    }
	}

	NodeList attribute_elements_nodelist = parent_element.getElementsByTagName(attribute_name);
	for (int i = 0; i < attribute_elements_nodelist.getLength(); i++) {
	    Element current_attribute_element = (Element) attribute_elements_nodelist.item(i);

	    // Check if the language matches
	    String current_attribute_element_language = current_attribute_element.getAttribute("language");
	    if (language_matches || current_attribute_element_language.equals(language_code)) {
		// Found attribute; returns its value
		return XMLTools.getElementTextValue(current_attribute_element);
	    }
	}

	return null;
    }


    public MetadataElement getMetadataElement(String metadata_element_name)
    {
	// Find the desired metadata element in the list
	for (int i = 0; i < metadata_set_elements.size(); i++) {
	    MetadataElement metadata_element = (MetadataElement) metadata_set_elements.get(i);
	    if (metadata_element.getName().equals(metadata_element_name)) {
		return metadata_element;
	    }
	}

	// Not found
	return null;
    }


    public ArrayList getMetadataSetElements()
    {
	return metadata_set_elements;
    }

   
    public String getNamespace()
    {
	return metadata_set_namespace;
    }
}
